import subprocess
import os
import pickle
import torch
import random
import time
import numpy as np
import pandas as pd
from cryptography.fernet import Fernet

# Import functions from model.py
from model import compile_oram, store, retrieve

# Simulate homomorphic encryption (for demonstration purposes)
def homomorphic_encrypt(data):
    # Placeholder for homomorphic encryption
    return data

def homomorphic_decrypt(data):
    # Placeholder for homomorphic decryption
    return data

def baseline_store(key, tensor):
    """Baseline storage without ORAM."""
    tensor_bytes = pickle.dumps(tensor)
    # Simulate storage (e.g., writing to a file or database)
    file_name = f"baseline_{key}.bin"
    with open(file_name, "wb") as f:
        f.write(tensor_bytes)
    return file_name

def baseline_retrieve(key):
    """Baseline retrieval without ORAM."""
    file_name = f"baseline_{key}.bin"
    with open(file_name, "rb") as f:
        tensor_bytes = f.read()
    return pickle.loads(tensor_bytes)

def homomorphic_store(key, tensor):
    """Store using homomorphic encryption."""
    tensor_bytes = pickle.dumps(tensor)
    encrypted_data = homomorphic_encrypt(tensor_bytes)
    file_name = f"homomorphic_{key}.bin"
    # Simulate storage
    with open(file_name, "wb") as f:
        f.write(encrypted_data)
    return file_name

def homomorphic_retrieve(key):
    """Retrieve using homomorphic encryption."""
    file_name = f"homomorphic_{key}.bin"
    with open(file_name, "rb") as f:
        encrypted_data = f.read()
    tensor_bytes = homomorphic_decrypt(encrypted_data)
    return pickle.loads(tensor_bytes)

def measure_latency_and_memory(store_func, retrieve_func, num_operations=100, is_oram=False):
    """Measures latency and memory overhead for storage and retrieval."""
    latencies = []
    memory_overhead = []

    for _ in range(num_operations):
        key = random.randint(0, 9)
        tensor = torch.randn(4, 4)

        # Measure storage latency
        start_time = time.time()
        if is_oram:
            store_func(key, tensor)
            # Estimate memory size for ORAM
            memory_size = len(pickle.dumps(tensor))
        else:
            file_name = store_func(key, tensor)
            memory_size = os.path.getsize(file_name)
        store_latency = time.time() - start_time

        # Measure retrieval latency
        start_time = time.time()
        retrieved_tensor = retrieve_func(key)
        retrieve_latency = time.time() - start_time

        latencies.append((store_latency, retrieve_latency))
        memory_overhead.append(memory_size)

    avg_store_latency = np.mean([lat[0] for lat in latencies])
    avg_retrieve_latency = np.mean([lat[1] for lat in latencies])
    avg_memory_overhead = np.mean(memory_overhead)

    return avg_store_latency, avg_retrieve_latency, avg_memory_overhead

def main():
    compile_oram()  # Ensure ORAM is compiled

    # Measure baseline
    baseline_results = measure_latency_and_memory(baseline_store, baseline_retrieve)

    # Measure Path ORAM
    path_oram_results = measure_latency_and_memory(store, retrieve, is_oram=True)

    # Measure Homomorphic Encryption
    homomorphic_results = measure_latency_and_memory(homomorphic_store, homomorphic_retrieve)

    # Compile results into a DataFrame
    results_df = pd.DataFrame({
        "Configuration": ["Baseline", "Path ORAM", "Homomorphic Encryption"],
        "Avg Store Latency (s)": [baseline_results[0], path_oram_results[0], homomorphic_results[0]],
        "Avg Retrieve Latency (s)": [baseline_results[1], path_oram_results[1], homomorphic_results[1]],
        "Avg Memory Overhead (bytes)": [baseline_results[2], path_oram_results[2], homomorphic_results[2]]
    })

    print(results_df)

if __name__ == "__main__":
    main()