import subprocess
import os
import pickle
import torch
import random
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from cryptography.fernet import Fernet
from scipy.stats import entropy

# Import functions from model.py
from model import compile_oram, store, retrieve

def calculate_entropy(data):
    """Calculates Shannon entropy of the data."""
    value, counts = np.unique(data, return_counts=True)
    return entropy(counts, base=2)

def calculate_kl_divergence(p, q):
    """Calculates KL divergence between two distributions."""
    return entropy(p, q, base=2)

def simulate_access_patterns(num_accesses=100):
    """Simulates access patterns and calculates entropy and KL divergence."""
    access_counts = np.zeros(10)  # Assuming 10 possible keys for simplicity
    for _ in range(num_accesses):
        key = random.randint(0, 9)
        store(key, torch.randn(4, 4))  # Store random data
        access_counts[key] += 1

    # Normalize access counts to get a probability distribution
    access_distribution = access_counts / num_accesses
    uniform_distribution = np.ones_like(access_distribution) / len(access_distribution)

    shannon_entropy = calculate_entropy(access_distribution)
    kl_divergence = calculate_kl_divergence(access_distribution, uniform_distribution)

    return shannon_entropy, kl_divergence, access_distribution

def plot_results(shannon_entropy, kl_divergence, access_distribution):
    """Displays the results of the entropy and KL divergence analysis as a table."""
    
    # Creating a DataFrame for tabular representation
    metrics_data = {
        "Metric": ["Shannon Entropy", "KL Divergence"],
        "Value": [shannon_entropy, kl_divergence]
    }
    df_metrics = pd.DataFrame(metrics_data)
    
    access_data = {
        "Key": list(range(len(access_distribution))),
        "Probability": access_distribution
    }
    df_access = pd.DataFrame(access_data)
    
    # Print tables to console
    print("\nEntropy and KL Divergence Metrics:")
    print(df_metrics.to_string(index=False))

    print("\nAccess Distribution:")
    print(df_access.to_string(index=False))

    # Optionally, display the table using matplotlib for a research-style presentation
    fig, ax = plt.subplots(1, 2, figsize=(12, 4))

    ax[0].axis('tight')
    ax[0].axis('off')
    ax[0].table(cellText=df_metrics.values, colLabels=df_metrics.columns, cellLoc='center', loc='center')

    ax[1].axis('tight')
    ax[1].axis('off')
    ax[1].table(cellText=df_access.values, colLabels=df_access.columns, cellLoc='center', loc='center')

    plt.suptitle("Research Results: Entropy, KL Divergence, and Access Distribution", fontsize=14, fontweight='bold')
    plt.show()

def main():
    compile_oram()  # Ensure ORAM is compiled

    shannon_entropy, kl_divergence, access_distribution = simulate_access_patterns()
    print(f"Shannon Entropy: {shannon_entropy}")
    print(f"KL Divergence: {kl_divergence}")

    plot_results(shannon_entropy, kl_divergence, access_distribution)

if __name__ == "__main__":
    main()