import subprocess
import pickle
import torch
from langchain_openai import ChatOpenAI
import hashlib
from cryptography.fernet import Fernet
import base64
import tempfile
import os
from transformers import AutoModelForCausalLM


# === Setup Encryption ===
ENCRYPTION_KEY = Fernet.generate_key()
cipher = Fernet(ENCRYPTION_KEY)


def compile_oram():
    """Compile ORAM C++ code."""
    print("Compiling ORAM...")
    subprocess.run(["make", "rebuild"], check=True)
    print("ORAM compiled.\n")


def hash_key(key):
    """Hash any key into SHA-256 hex string."""
    return hashlib.sha256(str(key).encode()).hexdigest()


def encrypt_data(data_bytes):
    """Encrypt raw bytes using Fernet."""
    return cipher.encrypt(data_bytes)


def decrypt_data(data_bytes):
    """Decrypt encrypted bytes."""
    return cipher.decrypt(data_bytes)


def store_to_oram(block_id, data_bytes):
    """Store encrypted data in ORAM using temp file."""
    print(f"Storing block_id: {block_id[:10]}... in ORAM")
    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        tmpfile.write(data_bytes)
        tmpfile_path = tmpfile.name
    subprocess.run(["./oram", "storefile", block_id, tmpfile_path], check=True)
    os.remove(tmpfile_path)


def retrieve_from_oram(block_id):
    """Retrieve and decrypt data from ORAM using temp file."""
    print(f"Retrieving block_id: {block_id[:10]}... from ORAM")
    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        tmpfile_path = tmpfile.name
    subprocess.run(["./oram", "retrievefile", block_id, tmpfile_path], check=True)
    with open(tmpfile_path, 'rb') as f:
        encrypted_bytes = f.read()
    os.remove(tmpfile_path)
    decrypted_bytes = decrypt_data(encrypted_bytes)
    return decrypted_bytes


def generate_response(query, api_key):
    """Generate LLM response."""
    chat_model = ChatOpenAI(model="gpt-4o-mini", openai_api_key=api_key)
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": query}
    ]
    response = chat_model.invoke(messages)
    return response


def main():
    compile_oram()

    # === 1. Load pretrained GPT-2 locally (proxy for GPT-4o-mini params) ===
    print("Loading GPT-2 model...")
    model = AutoModelForCausalLM.from_pretrained('gpt2')

    # === 2. Extract model parameters ===
    print("Extracting model parameters...")
    state_dict = model.state_dict()
    weights_bytes = b''
    for name, tensor in state_dict.items():
        weights_bytes += tensor.numpy().tobytes()
    print(f"Extracted {len(state_dict)} parameters, total size: {len(weights_bytes)} bytes\n")

    # === 3. Encrypt model parameters ===
    encrypted_weights = encrypt_data(weights_bytes)
    model_block_id = hash_key("gpt4o-mini-weights")

    # === 4. Store encrypted parameters in ORAM ===
    store_to_oram(model_block_id, encrypted_weights)

    # === 5. Retrieve & decrypt model parameters ===
    retrieved_bytes = retrieve_from_oram(model_block_id)
    print(f"Retrieved {len(retrieved_bytes)} bytes of model weights (decrypted successfully)\n")

    # === 6. Securely store & retrieve user query ===
    user_query = input("Enter Prompt: ")
    query_block_id = hash_key("user-query")
    encrypted_query = encrypt_data(user_query.encode())
    store_to_oram(query_block_id, encrypted_query)
    retrieved_query = retrieve_from_oram(query_block_id).decode()
    print(f"Retrieved Query: {retrieved_query}\n")

    # === 7. Generate response ===
    api_key = "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"
    response = generate_response(retrieved_query, api_key)
    print("AI Response:", response.content)


if __name__ == "__main__":
    main()
