/*
 * ORAM Implementation for Securing LLM Parameters
 * This code simulates Oblivious RAM (ORAM) to securely store and access model parameters.
 * ORAM ensures that access patterns do not leak sensitive information.
 * 
 * Author: Lily Faris
 */

#include <iostream>
#include <unordered_map>
#include <vector>
#include <random>
#include <fstream>
#include <algorithm>
#include <chrono>

using namespace std;

class ORAM {
private:
    unordered_map<int, string> storage;
    vector<int> accessHistory;
    hash<string> str_hash;
    const string filename = "oram_data.txt"; // Persistent storage file

    void load_from_file() {
        ifstream file(filename);
        if (file.is_open()) {
            int key;
            string value;
            while (file >> key) {
                file.ignore();
                getline(file, value);
                storage[key] = value;
            }
            file.close();
        }
    }

    void save_to_file() {
        ofstream file(filename, ios::trunc);
        if (file.is_open()) {
            for (const auto& pair : storage) {
                file << pair.first << " " << pair.second << "\n";
            }
            file.close();
        }
    }

public:
    ORAM() {
        load_from_file(); // Load stored parameters at initialization
    }
    
    void store(int key, string value) {
        size_t hashed_key = str_hash(to_string(key));
        storage[hashed_key] = value;
        accessHistory.push_back(hashed_key);
        save_to_file(); // Save changes
    }

    void perform_dummy_reads(int num_reads) {
        random_device rd;
        mt19937 gen(rd());
    
        uniform_int_distribution<int> dist(1, 100);  // Random key range
        uniform_int_distribution<int> noise_dist(1, 10);  // Random dummy read count
    
        int random_reads = noise_dist(gen);  // Randomly decide how many dummy reads to perform
    
        for (int i = 0; i < random_reads; i++) {  // Use random_reads instead of fixed 5
            int dummy_key = dist(gen);
            storage.find(str_hash(to_string(dummy_key))); // Perform fake access
        }
    }    

    void log_access(int key) {
        ofstream log_file("oram_access_log.txt", ios::app);
        if (log_file.is_open()) {
            log_file << "Accessed Key: " << key << " at " << chrono::system_clock::to_time_t(chrono::system_clock::now()) << "\n";
            log_file.close();
        }
    }
    
    string retrieve(int key) {
        size_t hashed_key = str_hash(to_string(key));

        ////////////////// THIS IS TO RANDOMIZED THE FREQUENCY OF THE RETREIEVED PARAMETER //////////////////
        //perform_dummy_reads(rand() % 10 + 1); // Random dummy reads between 1 and 10 to add noise to hide real accesses

        log_access(key); // Log every retrieval
        if (storage.find(hashed_key) != storage.end()) {
            return storage[hashed_key];
        }
        return "Parameter not found";
    }
    
    // Simulate ORAM shuffle to obfuscate access patterns
    void shuffle() {
        random_device rd;
        mt19937 g(rd());
        std::shuffle(accessHistory.begin(), accessHistory.end(), g);
    }

    // Display stored parameters (for testing only, not secure in real use)
    void debug_display() {
        cout << "Stored Parameters (hashed keys):\n";
        for (const auto& pair : storage) {
            cout << "Key: " << pair.first << " | Value: " << pair.second << endl;
        }
    }
};

int main(int argc, char* argv[]) {
    ORAM oram;

    if (argc < 2) {
        cout << "Usage: ./oram [store/retrieve/shuffle] [key] [value]" << endl;
        return 1;
    }

    string command = argv[1];

    if (command == "store" && argc == 4) {
        int key = stoi(argv[2]);
        oram.store(key, argv[3]);
        cout << "Stored successfully!" << endl;
    }
    else if (command == "retrieve" && argc == 3) {
        int key = stoi(argv[2]);
        cout << oram.retrieve(key) << endl;
    }
    else if (command == "shuffle") {
        cout << "Shuffling access patterns..." << endl;
        oram.shuffle();
        
        cout << "Displaying stored data (hashed keys)..." << endl;
        oram.debug_display();
    }
    else {
        cout << "Invalid command." << endl;
        return 1;
    }

    return 0;
}

