import subprocess
import os
from langchain_openai import ChatOpenAI
import random

def compile_oram():
    """Compiles the ORAM C++ code if not already compiled."""
    if not os.path.exists("oram"):  # Check if executable exists
        print("Compiling ORAM C++ code...")
        compile_command = ["g++", "oram.cpp", "-o", "oram"]
        subprocess.run(compile_command, check=True)
        print("Compilation complete.")

def store_parameter(key, value):
    """Stores a model parameter securely using ORAM."""
    subprocess.run(["./oram", "store", str(key), value], check=True)
    
def retrieve_parameter(key):
    """Retrieves a model parameter securely using ORAM."""
    result = subprocess.run(["./oram", "retrieve", str(key)], capture_output=True, text=True, check=True)
    return result.stdout.strip()

def perform_shuffle():
    """Calls the ORAM executable to shuffle access patterns."""
    subprocess.run(["./oram", "shuffle"], check=True)
  
def generate_response(query, api_key):
    """Generate a response using GPT-4o-mini without retrieval."""
    chat_model = ChatOpenAI(model="gpt-4o-mini", openai_api_key=api_key)
    
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": query}
    ]

    response = chat_model.invoke(messages)
    return response

def main():
    api_key = ""
    
    compile_oram()  # Ensure ORAM is compiled
    
    print("Storing model parameters...")
    parameters = {
        1: "weight_1: 0.345",
        2: "weight_2: 0.678",
        3: "bias_1: -0.123",
        4: "weight_3: 0.597",
        5: "bias_2: 0.214",
        6: "weight_4: -0.451"
    }
    
    n = len(parameters)

    for key, value in parameters.items():
        store_parameter(key, value)
    
    print("Retrieving stored parameters...")
    for key in parameters.keys():
        print(f"Retrieving key {key}:", retrieve_parameter(key)) 
        
    k = random.randint(n + 1, 10**9)
    print(f"Retrieved key {k}:", retrieve_parameter(k))  # This should return "Parameter not found"

    perform_shuffle() # Perform Oram Shuffle to randomize the access pattern
    
    query = "What is the capital of France and how is it?"
    response = generate_response(query, api_key)
    print("AI Response:", response.content)
    
if __name__ == "__main__":
    main()
