import subprocess
import os
from langchain_openai import ChatOpenAI
import time
import pandas as pd

access_log = []  # Stores timestamps and accessed keys

def compile_oram():
    """Compiles the ORAM C++ code if not already compiled."""
    if not os.path.exists("oram"):  # Check if executable exists
        print("Compiling ORAM C++ code...")
        compile_command = ["g++", "oram.cpp", "-o", "oram"]
        subprocess.run(compile_command, check=True)
        print("Compilation complete.")

def store_parameter(key, value):
    """Stores a model parameter securely using ORAM."""
    subprocess.run(["./oram", "store", str(key), value], check=True)

def analyze_logs(file_path):
    """Reads and analyzes ORAM access logs to detect patterns."""
    try:
        df = pd.read_csv(file_path, sep=" ", names=["Label", "Key", "At", "Timestamp"])
        print("\n--- Access Frequency Analysis ---")
        print(df["Key"].value_counts())  # Count frequency of access per key
        print("\n--- Sample Log Entries ---")
        print(df.head())  # Show first few logged accesses
    except Exception as e:
        print("Error analyzing logs:", e)

def retrieve_parameter(key):
    """Retrieves a model parameter (without ORAM) and logs access."""
    access_log.append((time.time(), key))  # Log timestamp and key
    result = subprocess.run(["./oram", "retrieve", str(key)], capture_output=True, text=True, check=True)
    return result.stdout.strip() if result.returncode == 0 else "Error unprotected-retrieving parameter"

def perform_shuffle():
    """Calls the ORAM executable to shuffle access patterns."""
    subprocess.run(["./oram", "shuffle"], check=True)

def generate_response(query, api_key):
    """Generate a response using GPT-4o-mini, obfuscating query parameters."""
    chat_model = ChatOpenAI(model="gpt-4o-mini", openai_api_key=api_key)

    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": query}
    ]

    response = chat_model.invoke(messages)

    # Randomly query dummy parameters
    dummy_keys = [random.randint(1, 10) for _ in range(3)]
    for key in dummy_keys:
        retrieve_parameter(key)  # Add fake retrievals

    return response

def main():
    """Main function to test ORAM behavior."""
    compile_oram()  # Ensure ORAM is compiled

    print("\n--- Storing Model Parameters in ORAM ---")
    parameters = {
        1: "weight_1: 0.345",
        2: "weight_2: 0.678",
        3: "bias_1: -0.123",
        4: "weight_3: 0.597"
    }

    for key, value in parameters.items():
        store_parameter(key, value)

    print("\n--- Retrieving Parameters Before ORAM (Unprotected) ---")
    for key in parameters.keys():
        print(f"Retrieving key {key}:", retrieve_parameter(key))

    print("\n--- Retrieving Parameters Using ORAM (Protected) ---")
    print("\n--- Performing ORAM Shuffle ---")
    perform_shuffle()
    
    # print("\n--- Retrieving Parameters Using ORAM (Protected) ---")
    # for key in parameters.keys():
    #     print(f"Retrieving key {key}:", retrieve_parameter(key))

    print("\n--- Analyzing ORAM Logs ---")
    analyze_logs("oram_access_log.txt")

if __name__ == "__main__":
    main()