// path_oram.cpp
// A simplified Path ORAM implementation using a tree-based structure,
// a position map, and a stash. This example is for demonstration only
// and is not secure for production use.

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unordered_map>
#include <algorithm>
#include <random>
#include <cstdlib>

using namespace std;

// ORAM parameters (for demonstration)
const int TREE_HEIGHT = 3;            // Height of the tree (number of levels - 1)
const int BUCKET_CAPACITY = 4;        // Maximum number of blocks per bucket
const int NUM_LEAVES = 1 << TREE_HEIGHT;             // 2^TREE_HEIGHT leaves
const int NUM_BUCKETS = (1 << (TREE_HEIGHT + 1)) - 1;  // Total number of buckets in a complete binary tree
const string STATE_FILE = "oram_state.db";           // File to persist ORAM state

// Structure for a data block.
struct Block {
    int id;         // Logical block identifier
    string data;    // Block data (e.g. hex-encoded string)
};

// Global ORAM state:
vector<vector<Block>> tree;              // Tree: vector of buckets (each bucket is a vector of blocks)
unordered_map<int, int> posMap;          // Position map: maps block id to assigned leaf (0 to NUM_LEAVES-1)
vector<Block> stash;                     // Stash: temporary storage for blocks during accesses

// Random number generator for new leaf assignments.
mt19937 rng(random_device{}());

// Helper: For a given leaf (0-indexed), return the indices (in the tree array) along the path from the root to that leaf.
vector<int> getPathIndices(int leaf) {
    int idx = leaf + (NUM_BUCKETS - NUM_LEAVES); // Map leaf number to bucket index in tree
    vector<int> path;
    while (true) {
        path.push_back(idx);
        if (idx == 0) break;
        idx = (idx - 1) / 2;
    }
    reverse(path.begin(), path.end());
    return path;
}

// Helper: Generate a random leaf number between 0 and NUM_LEAVES-1.
int getRandomLeaf() {
    uniform_int_distribution<int> dist(0, NUM_LEAVES - 1);
    return dist(rng);
}

// Load ORAM state from the persistent file. If the file is not found, initialize an empty state.
void loadState() {
    ifstream infile(STATE_FILE);
    if (!infile) {
        // No state file; initialize empty ORAM state.
        tree.assign(NUM_BUCKETS, vector<Block>());
        posMap.clear();
        stash.clear();
        return;
    }
    string line;
    // Expect first section: tree
    getline(infile, line);
    if(line != "#tree") {
        cerr << "State file corrupted: missing #tree" << endl;
        exit(1);
    }
    tree.clear();
    tree.resize(NUM_BUCKETS);
    for (int i = 0; i < NUM_BUCKETS; i++) {
        getline(infile, line);
        istringstream iss(line);
        int numBlocks;
        iss >> numBlocks;
        vector<Block> bucket;
        for (int j = 0; j < numBlocks; j++) {
            int id;
            string data;
            iss >> id >> data;
            bucket.push_back({id, data});
        }
        tree[i] = bucket;
    }
    // Next section: position map
    getline(infile, line);
    if(line != "#posmap") {
        cerr << "State file corrupted: missing #posmap" << endl;
        exit(1);
    }
    posMap.clear();
    int posEntries;
    infile >> posEntries;
    for (int i = 0; i < posEntries; i++) {
        int blockId, leaf;
        infile >> blockId >> leaf;
        posMap[blockId] = leaf;
    }
    infile.ignore(); // consume end-of-line
    // Next section: stash
    getline(infile, line);
    if(line != "#stash") {
        cerr << "State file corrupted: missing #stash" << endl;
        exit(1);
    }
    stash.clear();
    int stashEntries;
    infile >> stashEntries;
    for (int i = 0; i < stashEntries; i++) {
        int id;
        string data;
        infile >> id >> data;
        stash.push_back({id, data});
    }
    infile.close();
}

// Save ORAM state to the persistent file.
void saveState() {
    ofstream outfile(STATE_FILE, ios::trunc);
    // Save tree.
    outfile << "#tree" << endl;
    for (int i = 0; i < NUM_BUCKETS; i++) {
        outfile << tree[i].size();
        for (auto &blk : tree[i]) {
            outfile << " " << blk.id << " " << blk.data;
        }
        outfile << endl;
    }
    // Save position map.
    outfile << "#posmap" << endl;
    outfile << posMap.size() << endl;
    for (auto &p : posMap) {
        outfile << p.first << " " << p.second << endl;
    }
    // Save stash.
    outfile << "#stash" << endl;
    outfile << stash.size() << endl;
    for (auto &blk : stash) {
        outfile << blk.id << " " << blk.data << endl;
    }
    outfile.close();
}

// Eviction: Write blocks from the stash back into the tree along the given path.
// For each bucket on the path, try to place blocks from the stash that "belong" to that bucket.
void evictPath(const vector<int>& path) {
    for (int bucketIdx : path) {
        int capacity = BUCKET_CAPACITY - tree[bucketIdx].size();
        if (capacity <= 0) continue;
        // Iterate over stash and try to place blocks whose assigned leaf's path includes this bucket.
        for (auto it = stash.begin(); it != stash.end(); ) {
            int assignedLeaf = posMap[it->id]; // Each block’s current assigned leaf.
            vector<int> blockPath = getPathIndices(assignedLeaf);
            if (find(blockPath.begin(), blockPath.end(), bucketIdx) != blockPath.end()) {
                tree[bucketIdx].push_back(*it);
                it = stash.erase(it);
                capacity--;
                if (capacity == 0) break;
            } else {
                ++it;
            }
        }
    }
}

// The ORAM access operation.
// For a given block id, it reads the path from the tree (adding all blocks to the stash),
// updates the block (if writing), reassigns a new random leaf, and then evicts blocks back into the tree.
string accessBlock(int blockId, const string &newData, bool isWrite) {
    // Lookup block's current leaf in the position map; if not found, assign one.
    int oldLeaf;
    if (posMap.find(blockId) == posMap.end()) {
        oldLeaf = getRandomLeaf();
        posMap[blockId] = oldLeaf;
    } else {
        oldLeaf = posMap[blockId];
    }
    // Reassign the block a new random leaf.
    int newLeaf = getRandomLeaf();
    posMap[blockId] = newLeaf;
    
    // Determine the path from the root to the old leaf.
    vector<int> path = getPathIndices(oldLeaf);
    
    // Read the entire path: remove blocks from each bucket on the path and add them to the stash.
    for (int bucketIdx : path) {
        for (auto &blk : tree[bucketIdx]) {
            stash.push_back(blk);
        }
        tree[bucketIdx].clear();
    }
    
    // Search the stash for the block.
    bool found = false;
    string retData = "";
    for (auto &blk : stash) {
        if (blk.id == blockId) {
            if (isWrite) {
                blk.data = newData;
            }
            retData = blk.data;
            found = true;
            break;
        }
    }
    
    if (!found && isWrite) {
        // If the block does not exist, add it.
        Block newBlock;
        newBlock.id = blockId;
        newBlock.data = newData;
        stash.push_back(newBlock);
        retData = newData;
    }
    
    if (!found && !isWrite) {
        // For a read access, if the block isn’t found, report an error.
        return "Data not found";
    }
    
    // Evict blocks from the stash back into the buckets along the accessed path.
    evictPath(path);
    
    return retData;
}

int main(int argc, char* argv[]) {
    if (argc < 3) {
        cerr << "Usage: ./path_oram <command> <block_id> [data]" << endl;
        cerr << "Commands:" << endl;
        cerr << "  store <block_id> <data>   : Store or update data for block" << endl;
        cerr << "  retrieve <block_id>       : Retrieve data for block" << endl;
        return 1;
    }
    
    string command = argv[1];
    int blockId = stoi(argv[2]);
    
    // Load the current ORAM state.
    loadState();
    
    if (command == "store") {
        if (argc < 4) {
            cerr << "Usage: ./path_oram store <block_id> <data>" << endl;
            return 1;
        }
        string data = argv[3];
        string result = accessBlock(blockId, data, true);
        saveState();
        cout << "Stored block " << blockId << " with data: " << result << endl;
    } else if (command == "retrieve") {
        string result = accessBlock(blockId, "", false);
        saveState();
        cout << result;
    } else {
        cerr << "Unknown command: " << command << endl;
        return 1;
    }
    
    return 0;
}
