/*
    * Path ORAM Secure Storage for Model Parameters & Queries
    * Implements a tree-based ORAM with path eviction.
*/

#include <sstream>
#include <iomanip>
#include <iostream>
#include <vector>
#include <unordered_map>
#include <unordered_set>
#include <random>
#include <fstream>
#include <cmath>

using namespace std;

const int BUCKET_SIZE = 2; // Max blocks per bucket
const int TREE_HEIGHT = 4; // Determines ORAM tree depth
const int NUM_BUCKETS = (1 << (TREE_HEIGHT + 1)) - 1; // Total nodes in binary tree

struct Block {
    int key;
    string value;
    bool valid;
};

struct Bucket {
    Block blocks[BUCKET_SIZE];
};
 
class PathORAM {
private:
    vector<Bucket> tree;
    unordered_map<int, int> position_map; // Maps key to leaf position
    unordered_set<int> used_leaves;  // Track used leaves
    vector<Block> stash;
    random_device rd;
    mt19937 gen;


    int get_random_leaf() {
        int leaf;
        do {
            leaf = uniform_int_distribution<int>(1, (1 << TREE_HEIGHT) - 1)(gen);
        } while (used_leaves.find(leaf) != used_leaves.end());  // Ensure unique leaf
        used_leaves.insert(leaf);
        return leaf;
    }

    void read_path(int leaf, vector<Block>& retrieved_blocks) {
        cout << "DEBUG: Reading path ending at leaf " << leaf << endl;
        for (int i = 0; i <= TREE_HEIGHT; i++) {
            int node = leaf >> i;
            cout << "DEBUG: Reading from node " << node << endl;
            for (auto& block : tree[node].blocks) {
                if (block.valid) {
                    cout << "DEBUG: Retrieved Key = " << block.key << ", Value = " << block.value << endl;
                    retrieved_blocks.push_back(block);
                }
            }
        }
    }
    
    void write_path(int leaf) {
        cout << "DEBUG: Writing stash to path ending at leaf " << leaf << endl;
    
        for (int i = 0; i <= TREE_HEIGHT; i++) {
            int node = leaf >> i;
            cout << "DEBUG: Writing to node " << node << endl;
    
            // Only move at most 1 item from stash to this bucket
            if (!stash.empty()) {
                tree[node].blocks[0] = stash.back();  // Move one item to ORAM
                cout << "DEBUG: Evicting Key = " << stash.back().key << " to node " << node << endl;
                stash.pop_back();
            }
        }
    
        cout << "DEBUG: Stash size after write_path = " << stash.size() << endl;
    }    
    

    /*
    void read_path(int leaf, vector<Block>& retrieved_blocks) {
        for (int i = 0; i <= TREE_HEIGHT; i++) {
            int node = leaf >> i;
            for (auto& block : tree[node].blocks) {
                if (block.valid) {
                    retrieved_blocks.push_back(block);
                    block.valid = false; // Remove from ORAM
                }
            }
        }
    }
 
    void write_path(int leaf) {
        for (int i = 0; i <= TREE_HEIGHT; i++) {
            int node = leaf >> i;
            for (auto& block : tree[node].blocks) {
                if (!stash.empty()) {
                    block = stash.back();
                    stash.pop_back();
                } else {
                    block.valid = false;
                }
            }
        }
    }
    */

    // Function to convert string to hex
    string to_hex(const string& input) {
        stringstream hex_stream;
        for (unsigned char c : input) {
            hex_stream << hex << setw(2) << setfill('0') << (int)c;
        }
        return hex_stream.str();
    }

    void save_position_map() {
        unordered_map<int, int> temp_map;
    
        // Load existing position map first
        ifstream infile("position_map.txt");
        if (infile.is_open()) {
            int key, leaf;
            while (infile >> key >> leaf) {
                temp_map[key] = leaf;
            }
            infile.close();
        }
    
        // Merge existing map with the current position map
        for (const auto& pair : position_map) {
            temp_map[pair.first] = pair.second;
        }
    
        // Save merged map to file
        ofstream outfile("position_map.txt", ios::trunc);  // Open in truncate mode
        if (outfile.is_open()) {
            for (const auto& pair : temp_map) {
                outfile << pair.first << " " << pair.second << endl;
            }
            outfile.close();
            cout << "DEBUG: position_map saved with " << temp_map.size() << " entries.\n";
        } else {
            cerr << "Error: Could not save position_map to file.\n";
        }
    }
    

    void load_position_map() {
        position_map.clear();
        used_leaves.clear();  // Reset the used leaves set
    
        ifstream file("position_map.txt");
        if (file.is_open()) {
            int key, leaf;
            while (file >> key >> leaf) {
                position_map[key] = leaf;
                used_leaves.insert(leaf);  // Track already used leaves
            }
            file.close();
            cout << "DEBUG: Loaded " << position_map.size() << " keys into position_map from file.\n";
        } else {
            cerr << "Warning: Could not load position_map from file. It may not exist yet.\n";
        }
    }    
 
public:
    PathORAM() : gen(rd()) {
        tree.resize(NUM_BUCKETS);
        //load_position_map();
    }
    
    void store(int key, string value) {
        int leaf = get_random_leaf();
        position_map[key] = leaf;
    
        cout << "DEBUG: Stash size before adding key = " << stash.size() << endl;
        
        stash.push_back({key, value, true});
    
        cout << "DEBUG: Stash size after adding key = " << stash.size() << endl;
    
        // Only write to ORAM when stash reaches a threshold
        if (stash.size() >= 3) { // Evict only when stash has 3+ items
            write_path(leaf);
        }
    
        save_position_map();  // Save position map
    }     

    string retrieve(int key) {
        load_position_map();  // ✅ Load the position map from the file
    
        cout << "DEBUG: position_map at retrieval time:\n";
        for (const auto& pair : position_map) {
            cout << "Key: " << pair.first << " -> Leaf: " << pair.second << endl;
        }
    
        if (position_map.find(key) == position_map.end()) {
            cout << "Retrieve failed: Key " << key << " not found in position_map." << endl;
            return "Data not found";
        }
    
        int leaf = position_map[key];
        cout << "Retrieving Key: " << key << " from Leaf: " << leaf << endl;
    
        vector<Block> retrieved_blocks;
        read_path(leaf, retrieved_blocks);
    
        for (auto& block : retrieved_blocks) {
            cout << "Checking Block: Key = " << block.key << ", Value = " << block.value << endl;
            if (block.key == key) {
                return block.value;
            }
        }
    
        cout << "Retrieve failed: Key " << key << " not found in retrieved blocks." << endl;
        return "Data not found";
    }

    /*
    void store(int key, string value) {
        // int leaf = get_random_leaf();
        // position_map[key] = leaf;
    
        // string hex_value = to_hex(value);
        // stash.push_back({key, hex_value, true});
        // write_path(leaf);
    
        // // Debug print
        // cout << "Stored Key: " << key << " at Leaf: " << leaf << endl;
        int leaf = get_random_leaf();
        position_map[key] = leaf;
        stash.push_back({key, value, true});
        write_path(leaf);
        
        // Debug: Print the entire position map
        cout << "DEBUG: After storing, position_map contains:\n";
        for (const auto& pair : position_map) {
            cout << "Key: " << pair.first << " -> Leaf: " << pair.second << endl;
        }
    }

    string retrieve(int key) {
        cout << "DEBUG: position_map at retrieval time:\n";
        for (const auto& pair : position_map) {
            cout << "Key: " << pair.first << " -> Leaf: " << pair.second << endl;
        }
    
        if (position_map.find(key) == position_map.end()) {
            cout << "Retrieve failed: Key " << key << " not found in position_map." << endl;
            return "Data not found";
        }
    
        int leaf = position_map[key];
        cout << "Retrieving Key: " << key << " from Leaf: " << leaf << endl;
    
        vector<Block> retrieved_blocks;
        read_path(leaf, retrieved_blocks);
    
        for (auto& block : retrieved_blocks) {
            cout << "Checking Block: Key = " << block.key << ", Value = " << block.value << endl;
            if (block.key == key) {
                return block.value;
            }
        }
    
        cout << "Retrieve failed: Key " << key << " not found in retrieved blocks." << endl;
        return "Data not found";
    }
    */
    
    /*
    string retrieve(int key) {
        cout << "Current position_map:\n";
        for (const auto& pair : position_map) {
            cout << "Key: " << pair.first << " -> Leaf: " << pair.second << endl;
        }
    
        if (position_map.find(key) == position_map.end()) {
            cout << "Retrieve failed: Key " << key << " not found in position_map." << endl;
            return "Data not found";
        }
    
        int leaf = position_map[key];
        cout << "Retrieving Key: " << key << " from Leaf: " << leaf << endl;
    
        vector<Block> retrieved_blocks;
        read_path(leaf, retrieved_blocks);
    
        for (auto& block : retrieved_blocks) {
            cout << "Checking Block: Key = " << block.key << ", Value = " << block.value << endl;
            if (block.key == key) {
                return block.value;
            }
        }
    
        cout << "Retrieve failed: Key " << key << " not found in retrieved blocks." << endl;
        return "Data not found";
    }    

    
    string retrieve(int key) {
        if (position_map.find(key) == position_map.end()) {
            cout << "Retrieve failed: Key " << key << " not found in position_map." << endl;
            return "Data not found";
        }
    
        int leaf = position_map[key];
        cout << "Retrieving Key: " << key << " from Leaf: " << leaf << endl;
    
        vector<Block> retrieved_blocks;
        read_path(leaf, retrieved_blocks);
    
        for (auto& block : retrieved_blocks) {
            cout << "Checking Block: Key = " << block.key << ", Value = " << block.value << endl;
            if (block.key == key) {
                return block.value;
            }
        }
    
        cout << "Retrieve failed: Key " << key << " not found in retrieved blocks." << endl;
        return "Data not found";
    }
    */

    /*
    void store(int key, string value) {
        int leaf = get_random_leaf();
        position_map[key] = leaf;
        
        // Convert the value to hex format
        string hex_value = to_hex(value);

        stash.push_back({key, hex_value, true});
        write_path(leaf);
        cout << "stored successfully";
    }

    string retrieve(int key) {
        if (position_map.find(key) == position_map.end()) {
            return "Data not found";
        }
        int leaf = position_map[key];
        vector<Block> retrieved_blocks;
        read_path(leaf, retrieved_blocks);
        
        for (auto& block : retrieved_blocks) {
            if (block.key == key) {
                cout << block.value << endl; // Ensure it prints HEX
                return block.value;  // Now correctly returning HEX string
            }
        }
        return "Data not found";
    }   
    */ 

    /*
    void store(int key, string value) {
        int leaf = get_random_leaf();
        position_map[key] = leaf;
        stash.push_back({key, value, true});
        write_path(leaf);
    }

    string retrieve(int key) {
        if (position_map.find(key) == position_map.end()) {
            return "Data not found";
        }
        int leaf = position_map[key];
        vector<Block> retrieved_blocks;
        read_path(leaf, retrieved_blocks);
        for (auto& block : retrieved_blocks) {
            if (block.key == key) {
                return block.value;
            }
        }
        return "Data not found";
    }
    */

};
 
int main(int argc, char* argv[]) {
    PathORAM oram;
    if (argc < 2) {
        cout << "Usage: ./path_oram [store/retrieve] [key] [value]" << endl;
        return 1;
    }
    string command = argv[1];
    if (command == "store" && argc == 4) {
        int key = stoi(argv[2]);
        oram.store(key, argv[3]);
    } else if (command == "retrieve" && argc == 3) {
        int key = stoi(argv[2]);
        cout << oram.retrieve(key) << endl;
    } else {
        cout << "Invalid command." << endl;
        return 1;
    }
}
 