/*
 * ORAM Implementation for Securely Storing Queries & Model Parameters
 * This version supports query storage and retrieval with obfuscated access patterns.
 *
 * Author: Lily Faris
 */

 #include <iostream>
 #include <unordered_map>
 #include <vector>
 #include <random>
 #include <fstream>
 #include <algorithm>
 #include <chrono>
 
 using namespace std;
 
 class ORAM {
 private:
    unordered_map<int, string> storage; // Securely stores parameters & queries
    vector<int> accessHistory; // Tracks access patterns
    hash<string> str_hash;
    const string filename = "oram_data.txt"; // Persistent storage file

    void load_from_file() {
        ifstream file(filename);
        if (file.is_open()) {
            int key;
            string value;
            while (file >> key) {
                file.ignore();
                getline(file, value);
                storage[key] = value;
            }
            file.close();
        }
    }
 
     void save_to_file() {
        ofstream file(filename, ios::trunc);
        if (file.is_open()) {
            for (const auto& pair : storage) {
                file << pair.first << " " << pair.second << "\n";
            }
            file << "\n";
            file.close();
        }
     }
 
 public:
    ORAM() {
        load_from_file(); // Load stored data at initialization
    }
    
    void store(int key, string value) {
    
        storage[key] = value;
        accessHistory.push_back(key);

        size_t hashed_key = str_hash(to_string(key));
        storage[hashed_key] = value;
        accessHistory.push_back(hashed_key);

        save_to_file();
    }
     
     /*
    void perform_dummy_reads() {
        random_device rd;
        mt19937 gen(rd());
        uniform_int_distribution<int> dist(1, 100);
        for (int i = 0; i < rand() % 5 + 1; i++) {
            int dummy_key = dist(gen);
            storage.find(str_hash(to_string(dummy_key))); // Fake access
        }
    }
    */
 
    void log_access(int key) {
        ofstream log_file("oram_access_log.txt", ios::app);
        if (log_file.is_open()) {
            log_file << "Accessed Key: " << key << " at "
                    << chrono::system_clock::to_time_t(chrono::system_clock::now()) << "\n";
            log_file.close();
        }
    }
     
    string retrieve(int key) {
        size_t hashed_key = str_hash(to_string(key));
        //perform_dummy_reads(); // Add noise but its not working i don't think
        log_access(key);
        if (storage.find(hashed_key) != storage.end()) {
            return storage[hashed_key];
        }
        return "Data not found";
    }
     
    void shuffle() {
        random_device rd;
        mt19937 g(rd());
        std::shuffle(accessHistory.begin(), accessHistory.end(), g);
    }

    void debug_display() {
        ofstream shuffle_file("shuffling.txt", ios::trunc);
        if (shuffle_file.is_open()) {
            shuffle_file << "Stored Parameters:\n";
            for (const auto& pair : storage) {
                shuffle_file << "Key: " << pair.first << " | Value: " << pair.second << endl;
            }
            shuffle_file.close();
        }
    }
 };
 
int main(int argc, char* argv[]) {
    ORAM oram;

    if (argc < 2) {
        cout << "Usage: ./oram [store/retrieve/shuffle] [key] [value]" << endl;
        return 1;
    }

    string command = argv[1];

    if (command == "store" && argc == 4) {
        int key = stoi(argv[2]);
        oram.store(key, argv[3]);
    }
    else if (command == "retrieve" && argc == 3) {
        int key = stoi(argv[2]);
        cout << oram.retrieve(key) << endl;
    }
    else if (command == "shuffle") {
        cout << "Shuffling access patterns..." << endl;
        oram.shuffle();
        oram.debug_display();
    }
    else {
        cout << "Invalid command." << endl;
        return 1;
    }
}