import subprocess
import os
import pickle
import torch  # For handling model parameters
from langchain_openai import ChatOpenAI
import random

def compile_oram():
    """Compiles the ORAM C++ code if not already compiled."""
    if not os.path.exists("test"):  # Check if executable exists
        print("Compiling ORAM C++ code...")
        compile_command = ["g++", "test.cpp", "-o", "test"]
        subprocess.run(compile_command, check=True)
        print("Compilation complete.")

def store(key, tensor):
    """Stores a PyTorch tensor securely using ORAM."""
    tensor_bytes = pickle.dumps(tensor)  # Serialize tensor
    subprocess.run(["./test", "store", str(key), tensor_bytes.hex()], check=True)

    
def retrieve(key):
    """Retrieves a value securely from ORAM."""
    result = subprocess.run(["./test", "retrieve", str(key)], capture_output=True, text=True, check=True)
    tensor_bytes = bytes.fromhex(result.stdout.strip())
    return pickle.loads(tensor_bytes)

'''
def shuffle():
    """Calls the ORAM executable to shuffle access patterns."""
    subprocess.run(["./test", "shuffle"], check=True)
'''
def generate_response(query, api_key):
    """Generate a response using GPT-4o-mini."""
    chat_model = ChatOpenAI(model="gpt-4o-mini", openai_api_key=api_key)
    
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": query}
    ]
    
    response = chat_model.invoke(messages)
    return response

def main():
    
    compile_oram()  # Ensure ORAM is compiled
    
    parameters = {
        1: torch.randn(4, 4),  # Simulated weight matrix
        2: torch.randn(4),      # Simulated bias vector
        3: torch.randn(4, 4),
        4: torch.randn(4),
    }
    
    # Store and retrieve user queries securely
    user_query = input("Enter Prompt here: ")
    api_key = ""
    query_key = random.randint(len(parameters), 10000)  # Assign a random key
    
    print("Storing initial model parameters...")
    store(query_key, user_query)
    for key, tensor in parameters.items():
        store(key, tensor)
        
    #shuffle()  # Shuffle ORAM access patterns

    print("Retrieving stored parameters...")
    for key in parameters.keys():
        print(f"Retrieving key {key}:", retrieve(key))
    
    retrieved_query = retrieve(query_key)
    print(f"Retrieved Query: {retrieved_query}")
    
    response = generate_response(retrieved_query, api_key)
    print("AI Response:", response.content)
    
    
if __name__ == "__main__":
    main()