#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>
#include <string>
#include <unordered_map>
#include <algorithm>
#include <random>
#include <cstdlib>

using namespace std;

// ORAM parameters (for demonstration)
const int TREE_HEIGHT = 5;            // Height of the tree (number of levels - 1)
const int BUCKET_CAPACITY = 5;        // Maximum number of blocks per bucket
const int NUM_LEAVES = 1 << TREE_HEIGHT;             // 2^TREE_HEIGHT leaves
const int NUM_BUCKETS = (1 << (TREE_HEIGHT + 1)) - 1;  // Total number of buckets in a complete binary tree
const string STATE_FILE = "oram_state.db";           // File to persist ORAM state


// Structure for a data block.
struct Block {
    string id;     // Hashed block identifier (SHA-256 hex string)
    string data;   // Block data (e.g. hex-encoded string)
};


// Global ORAM state:
vector<vector<Block>> tree;              // Tree: vector of buckets (each bucket is a vector of blocks)
unordered_map<string, int> posMap;       // Position map: maps block id (hashed string) to assigned leaf
vector<Block> stash;                     // Stash: temporary storage for blocks during accesses

// Random number generator for new leaf assignments.
mt19937 rng(random_device{}());

// Helper: path indices from root to leaf
vector<int> getPathIndices(int leaf) {
    int idx = leaf + (NUM_BUCKETS - NUM_LEAVES);
    vector<int> path;
    while (true) {
        path.push_back(idx);
        if (idx == 0) break;
        idx = (idx - 1) / 2;
    }
    //reverse(path.begin(), path.end());
    return path;
}

// Helper: Generate a random leaf number between 0 and NUM_LEAVES-1.
int getRandomLeaf() {
    uniform_int_distribution<int> dist(0, NUM_LEAVES - 1);
    return dist(rng);
}

// Load ORAM state firm the persistent file.
void loadState() {
    ifstream infile(STATE_FILE);
    if (!infile) {
        tree.assign(NUM_BUCKETS, vector<Block>());
        posMap.clear();
        stash.clear();
        return;
    }

    string line;
    getline(infile, line);
    if (line != "#tree") {
        cerr << "State file corrupted: missing #tree" << endl;
        exit(1);
    }

    tree.clear();
    tree.resize(NUM_BUCKETS);
    for (int i = 0; i < NUM_BUCKETS; i++) {
        getline(infile, line);
        istringstream iss(line);
        int numBlocks;
        iss >> numBlocks;
        vector<Block> bucket;
        for (int j = 0; j < numBlocks; j++) {
            string id, data;
            iss >> id >> data;
            bucket.push_back({id, data});
        }
        tree[i] = bucket;
    }

    getline(infile, line);
    if (line != "#posmap") {
        cerr << "State file corrupted: missing #posmap" << endl;
        exit(1);
    }

    posMap.clear();
    int posEntries;
    infile >> posEntries;
    for (int i = 0; i < posEntries; i++) {
        string blockId;
        int leaf;
        infile >> blockId >> leaf;
        posMap[blockId] = leaf;
    }
    infile.ignore();

    getline(infile, line);
    if (line != "#stash") {
        cerr << "State file corrupted: missing #stash" << endl;
        exit(1);
    }

    stash.clear();
    int stashEntries;
    infile >> stashEntries;
    for (int i = 0; i < stashEntries; i++) {
        string id, data;
        infile >> id >> data;
        stash.push_back({id, data});
    }
    infile.close();
}

// Save ORAM state frim the persistent file
void saveState() {
    ofstream outfile(STATE_FILE, ios::trunc);
    outfile << "#tree" << endl;
    for (int i = 0; i < NUM_BUCKETS; i++) {
        outfile << tree[i].size();
        for (auto &blk : tree[i]) {
            outfile << " " << blk.id << " " << blk.data;
        }
        outfile << endl;
    }

    outfile << "#posmap" << endl;
    outfile << posMap.size() << endl;
    for (auto &p : posMap) {
        outfile << p.first << " " << p.second << endl;
    }

    outfile << "#stash" << endl;
    outfile << stash.size() << endl;
    for (auto &blk : stash) {
        outfile << blk.id << " " << blk.data << endl;
    }
    outfile.close();
}

// Eviction function with dummy block insertion.
void evictPath(vector<int> path) {
    reverse(path.begin(), path.end());
    for (int bucketIdx : path) {
        int capacity = BUCKET_CAPACITY - tree[bucketIdx].size();
        if (capacity <= 0) continue;

        vector<string> toRemove;
        for (auto it = stash.begin(); it != stash.end(); ++it) {
            int assignedLeaf = posMap[it->id];
            vector<int> blockPath = getPathIndices(assignedLeaf);
            if (find(blockPath.begin(), blockPath.end(), bucketIdx) != blockPath.end()) {
                tree[bucketIdx].push_back(*it);
                toRemove.push_back(it->id);
                if (--capacity == 0) break;
            }
        }

        while (capacity > 0) {
            Block dummyBlock;
            dummyBlock.id = "DUMMY";
            dummyBlock.data = "dummy";
            tree[bucketIdx].push_back(dummyBlock);
            --capacity;
        }

        for (const string& id : toRemove) {
            stash.erase(remove_if(stash.begin(), stash.end(), [&](Block b) { return b.id == id; }), stash.end());
        }
    }
}

// ORAM access function.
string accessBlock(const string &blockId, const string &newData, bool isWrite) {
    int oldLeaf;
    if (posMap.find(blockId) == posMap.end()) {
        oldLeaf = getRandomLeaf();
        posMap[blockId] = oldLeaf;
    } else {
        oldLeaf = posMap[blockId];
    }

    int newLeaf = getRandomLeaf();
    posMap[blockId] = newLeaf;

    vector<int> path = getPathIndices(oldLeaf);

    for (int bucketIdx : path) {
        for (auto &blk : tree[bucketIdx]) {
            stash.push_back(blk);
        }
        tree[bucketIdx].clear();
    }

    bool found = false;
    string retData = "";
    for (auto &blk : stash) {
        if (blk.id == blockId) {
            if (isWrite) blk.data = newData;
            retData = blk.data;
            found = true;
            break;
        }
    }

    if (!found && isWrite) {
        stash.push_back({blockId, newData});
        retData = newData;
    }

    if (!found && !isWrite) return "Data not found";

    evictPath(path);

    return retData;
}

// Main
int main(int argc, char* argv[]) {
    if (argc < 3) {
        cerr << "Usage: ./path_oram <command> <hashed_key> [data/filepath]" << endl;
        return 1;
    }

    string command = argv[1];
    string blockId = argv[2];

    loadState();

    if (command == "store" && argc == 4) {
        cout << accessBlock(blockId, argv[3], true) << endl;
    } else if (command == "retrieve") {
        cout << accessBlock(blockId, "", false) << endl;
    } else if (command == "storefile" && argc == 4) {
        string filepath = argv[3];
        ifstream infile(filepath, ios::binary);
        if (!infile) {
            cerr << "Failed to open file for reading: " << filepath << endl;
            return 1;
        }
        stringstream buffer;
        buffer << infile.rdbuf();
        string data = buffer.str();
        infile.close();
        accessBlock(blockId, data, true);
        cout << "File stored successfully." << endl;
    } else if (command == "retrievefile" && argc == 4) {
        string filepath = argv[3];
        string data = accessBlock(blockId, "", false);
        if (data == "Data not found") {
            cerr << "Data not found for block ID: " << blockId << endl;
            return 1;
        }
        ofstream outfile(filepath, ios::binary);
        if (!outfile) {
            cerr << "Failed to open file for writing: " << filepath << endl;
            return 1;
        }
        outfile << data;
        outfile.close();
        cout << "File retrieved successfully." << endl;
    } else {
        cerr << "Invalid command" << endl;
        return 1;
    }

    saveState();
    return 0;
}
